#!/usr/bin/python3
from gpiozero import DigitalOutputDevice
from time import sleep

class Out74595:
    """Control a sequence of 74HC595 shift registers for output.

    Connection of 74HC595
    ---------------------
    -Reset (10)         -> high
    -Out enable (13)    -> low (or otherwise, if needed)
    Data in (14)        -> raspi pin_data
    Clock shift (11)    -> raspi pin_clk_shift
    Clock latch (12)    -> raspi pin_clk_latch


    Any number of 74HC595 shift registers can be cascaded by connecting output
    Q7' (pin 9) to data input (pin 14) of the subsequent shift register. The
    lines serial clock (pin 11) and latch clock (pin 12) of all shift
    registers are connected together.
    """

    # sleeping time 10 us
    tic_time = 10e-6

    def __init__(self, pin_data, pin_clk_shift, pin_clk_latch, cascade=1):
        """
            pin_data : string or int
                raspi pin used to control data input of 74HC595 (pin 14)

            pin_clk_shift : string or int
                raspi pin used to control serial clock input of 74HC595 (pin 11)

            pin_clk_latch : string or int
                raspi pin used to control latch clock input of 74HC595 (pin 12)

            cascade : int
                number of cascaded 74HC595 devices
        """
        self.pin_data = DigitalOutputDevice(pin_data)
        self.pin_clk_shift = DigitalOutputDevice(pin_clk_shift)
        self.pin_clk_latch = DigitalOutputDevice(pin_clk_latch)
        self.cascade = cascade

    def set(self, x):
        """Set outputs of shift registers to value x.

        x : list of boolean
            x must have exactly 8 * cascade elements.

            The first element is mapped to Q7 of the last shift register, the
            last element is mapped to Q0 of the first shift register.
        """
        if len(x) != self.cascade * 8:
            raise ValueError("Out74595.set(): list of %d elements needed" 
                % (self.cascade * 8))
        self.pin_clk_latch.off()
        for b in x:
            self.pin_clk_shift.off()
            self.pin_data.value = b
            sleep(self.tic_time)
            self.pin_clk_shift.on()
            sleep(self.tic_time)
        self.pin_clk_latch.on()

    def clear(self, clear_value=False):
        """Swich all outputs to clear_value (default = False)."""
        x = [clear_value for i in range(self.cascade * 8)]
        self.set(x)


# ----------------------------------------------------------------------------
# Test: Cascade of two 74HC595 shift registers
# ----------------------------------------------------------------------------

if __name__ == "__main__":
    my_register = Out74595(10, 11, 9, cascade=2)
    while True:
        for i in range(2 * 8):
            x = [i==j for j in range(2 * 8)]
            my_register.set(x)
            sleep(0.1)


# vim: set et:
