#!/usr/bin/python3
from ovf_client_common import OVF_Client
from gpiozero import Button
from incremental_encoder import Incremental_Encoder
from out74595 import Out74595
from signal import pause, signal, SIGINT, SIGTERM
from time import sleep

# ----------------------------------------------------------------------------
# Configuration
# ----------------------------------------------------------------------------

# Pulseaudio config
#
# Source (Microphone)
# Sink (Speaker)
#
# Usually, on a raspy the correct pulseaudio source and sink can be guessed
# automatically. However, if the guessing does not work (e. g. due to
# unexpected names), the correct sink and/or source can be given here as a
# string. To get a list of available sinks or sources run the commands
#
#   pactl list sources short | grep -v monitor | cut -f 2
#   pactl list sinks short | cut -f 2
#
# The RasPi's built-in sink has the term 'bcm...' (meaning Broadcom) in its
# name.
pa_source = None
pa_sink = None
pa_source_volume = 100
pa_sink_volume = 77

# GPIO pinout

# LED shift register
led_pin_data = 'GPIO10'
led_pin_shift = 'GPIO11'
led_pin_latch = 'GPIO9'

# incremental encoder (volume knob)
ie_pin_A = 'GPIO6'
ie_pin_B = 'GPIO5'

# push buttons
btn_pin_listen = 'GPIO13'
btn_pin_int = 'GPIO19'
btn_pin_ptt = 'GPIO26'
btn_bounce_time = 3e-3 # 3 ms

# ----------------------------------------------------------------------------
# Main program
# ----------------------------------------------------------------------------

vol_bargraph_max = 9
vol_default = pa_sink_volume

class UI:
    def __init__(self):
        # create ui elements
        self.ie = Incremental_Encoder(ie_pin_A, ie_pin_B)
        self.btn_listen = Button(btn_pin_listen, bounce_time=btn_bounce_time)
        self.btn_int = Button(btn_pin_int, bounce_time=btn_bounce_time)
        self.btn_ptt = Button(btn_pin_ptt, bounce_time=btn_bounce_time)
        self.leds = Out74595(led_pin_data, led_pin_shift, led_pin_latch,
                             cascade=2)

        # create and start OVF client
        self.ovf = OVF_Client(self.ui_update_callback,
                             # Pulseaudio
                             pa_source, pa_sink,
                             pa_source_volume, pa_sink_volume)

        # connect buttons
        self.ie.on_turn = self.turn_vol_knob
        self.btn_listen.when_pressed = self.ovf.ui_ev_press_listen
        self.btn_int.when_pressed = self.ovf.ui_ev_press_int
        self.btn_ptt.when_pressed = self.ovf.ui_ev_PTT_press
        self.btn_ptt.when_released = self.ovf.ui_ev_PTT_release

        # initialize values
        self._indicator_leds = [True for i in range(5)] # all five leds off
        self._vol_bargraph = self._vol_to_bargraph(vol_default)

        # initialize leds
        self.leds.clear(clear_value=False) # all on
        sleep(1)
        self.leds.clear(clear_value=True) # all off
        sleep(1)
        self.ui_update()

    def run(self):
        self.ovf.start()
        # wait for event (usually forever)
        # SIGINT and SIGTERM shall be handled gracefully and end the pause.
        pause()

    def done(self):
        self.leds.clear(clear_value=True)
        # call done() for all childeren implementing it
        self.ovf.done()

    # --- volume -------------------------------------------------------------

    def turn_vol_knob(self, direction):
        self.ovf.ui_change_volume(direction * int(100/vol_bargraph_max))

    def _vol_to_bargraph(self, vol):
        "Convert vol from [0..100] to nuber in [0..9]"
        return int(round(vol/100*9))

    # --- display update -----------------------------------------------------
    #
    # Set LED bits: 
    #   False, 0: on
    #   True,  1: off
    #
    # There are 16 led bits:
    # 00, 01, 02, 03, 04, 05, 06, 07, 08, 09, 10, 11, 12, 13, 14, 15
    # |low -------- volume ------- high|  NC  NC |--- indicator ---|
    #                                            con rx  tx lstn int

    def ui_update(self):
        # set 9 bits according to volume
        bargraph = [i >= self._vol_bargraph for i in range(9)]
        not_connected = [True, True]
        self.leds.set(bargraph + not_connected + self._indicator_leds)

    def ui_update_callback(self, RX, TX, Listen, Int, Con, Vol):
        """Callback for OVF_Client"""
        # invert led state (on -> False)
        self._indicator_leds = [not Con, not RX, not TX, not Listen, not Int]
        self._vol_bargraph = self._vol_to_bargraph(Vol)
        self.ui_update()

# ----------------------------------------------------------------------------
# main program
# ----------------------------------------------------------------------------

# signal handler do ignore signals
def sig_handler(signum, stack):
    """Signalhandler: Print signal name and do nothing."""
    print("Caught signal '%s'." % str(signum))

# install signal handlers
signal(SIGINT, sig_handler)
signal(SIGTERM, sig_handler)

# run main program
my_ui = UI()
try:
    my_ui.run()
finally:
    my_ui.done()

# vim: set et:
