#!/usr/bin/python3
import subprocess as sp

verbose = True

# ----------------------------------------------------------------------------
# Audiostream: RTP via pulseaudio
# ----------------------------------------------------------------------------
#
# The audio of the TRX and all clients is broadcasted in one RTP stream
# serving as a virtual ether.
#
# Set-up of pulseaudio modules
# ----------------------------
#
#  rtp-recv         null-sink       loopback    speaker sink
#                   * name          * rx
#                     rtp-rx-sink
#   ,----,,        ,,------,        ,----,,      ,,------,
#  /      \\       ||      |       /      \\     ||      |
# (       ||==>====||      |-->----|      ||==>==||      |
#  \      //       ||      |       \      //     ||      |
#   '----''        ''------'        '----''      ''------'
#
#  rtp-send         null-sink       loopback    mic/line-in
#                   * name          * tx        source
#                     rtp-tx-sink
#   ,-----,        ,------,,        ,,----,      ,------,
#  /       \       |      ||       //      \     |       \
# (        |--<----|      ||==<====||      |--<--|        )
#  \       /       |      ||       \\      /     |       /
#   '-----'        '------''        ''----'      '------'
#
# The null-sinks rtp-tx-sink and rtp-rx-sink are used to mute the streams. In
# this way, other applications using the same speaker/mic are not disturbed.
# Furthermore, the FFT-audio-thread can connect to the mic-source
# independently without being muted.
#
#
# Legend of symbols
# -----------------
# {{{ Doku
#
# Streams:
# ===== writing stream (data is pushed by player)
# ----- reading stream (data is pulled by recorder)
#
# Sink:                                 Source:
# || multiple writeable                 | multiple readable
# |  multiple readable (as .monitor)
#      ,,------,                             ,------,
#      ||      |                             |       \
# ==>==||      |-->--                   --<--|        )
#      ||      |                             |       /
#      ''------'                             '------'
#
# Playing Module/Program:               Redording Module/Program:
# || sigle writing                      | single reading
#       ,,----,                               ,-----,
#      //      \                             /       \
# ==<==||       )                       -->--|        )
#      \\      /                             \       /
#       ''----'                               '-----'
#
# Loopback Module                       Combined sink:
# || single writing                     || multiple writable
# |  single reading                     || multiple writing
#       ,,----,                              ,,-----,,
#      //      \                             ||     ||
# ==<==||      |--<--                   ==>==||     ||==>==
#      \\      /                             ||     ||
#       ''----'                              ''-----''
# }}}
# ----------------------------------------------------------------------------

class PA_RTP:
    """RTP sender and receiver.

    Both signal paths can be muted independently.
        tx_mute(bool)       mute/unmute rtp-tx path
        rx_mute(bool)       mute/unmute rtp-rx path
        set_volume(vol)     set volume of rx path in %

    On startup, the mute state of both path seems to be undefined. They must
    be initiallised correcrly by the calling instance.

    pa_source   audio input to read from
    pa_sink     audio output to write to
        If these paramters a omitted or 'None', the correct devices are
        guessed. This works only on a RasPi with exatly one USB sound card.
    """

    def __init__(self,
                 rtp_mcast_ip='224.0.0.47',
                 pa_source=None,
                 pa_sink=None,
                 pa_source_volume=50,
                 pa_sink_volume=50):

        # Pulseaudio network parameters
        self.rtp_mcast_ip = rtp_mcast_ip

        # Pulseaudio module numbers of rtp send and recv module
        self.module_rtp_recv = None
        self.module_rtp_send = None
        self.module_rtp_rx_sink = None
        self.module_rtp_tx_sink = None
        self.module_loopback_rx = None
        self.module_loopback_tx = None

        # {{{ PulseAudio documentation
        #
        # PulseAudio Modules are documented at <https://www.freedesktop.org/
        #   wiki/Software/PulseAudio/Documentation/User/Modules/>
        #
        # Additional documentation at:
        # <https://www.freedesktop.org/wiki/Software/PulseAudio/Notes/13.0/>
        #
        #   New "max_latency_msec" module argument for module-loopback
        #
        #   The "max_latency_msec" module argument can be used to set an upper
        #   limit for the loopback latency. By default the latency will
        #   automatically increase if there are underruns, but if keeping the
        #   latency below a certain limit is more important than the absence
        #   of glitches, then this option is useful.
        #
	#   DL1STO: Experience: If max_latency_msec is too small (10 ms), the
	#   voice seems to be pitched up (schlumpfig). Thus the value is set to
	#   60 ms.
        #
        #   New "fast_adjust_threshold_msec" module argument for
        #   module-loopback 
        #
        #   If the loopback for some reason gathers a lot of audio in its
        #   internal queue, it will take a long time before the loopback
        #   gradually adjusts its latency back to the target latency. This can
        #   happen for example during a suspend/resume cycle, if the
        #   loopback's output device takes more time to initialize than the
        #   input device. The new "fast_adjust_threshold_msec" module argument
        #   can be used to specify a threshold of difference between the
        #   actual latency and the target latency, and if the threshold is
        #   exceeded, the loopback immediately adjusts to the target latency
        #   by dropping audio from its queue.
        #
        # Additional additional documentation gathered from PulseAudio source
        # code:
        #
        #   module-loopback:
        #       fast_adjust_threshold_msec: value must be >= 100
        #
        # Messages from PulseAudio are written to /var/log/syslog.
        # }}}

        # Start pulseaudio
        exit_status = sp.run(['pulseaudio', '--check'])
        if exit_status.returncode == 0:
            self.pa_was_running = True
            if verbose: print("Pulseaudio already running")
        else:
            self.pa_was_running = False
            sp.run(['pulseaudio', '--start', '--daemonize', '--realtime'])
        exit_status = sp.run(['pulseaudio', '--check'])
        if exit_status.returncode != 0:
            raise RuntimeError("Cannot start pulseaudio -- exit")

        # Pulseaudio sinks and sources
        self.pa_sink = pa_sink if pa_sink else guess_sink()
        self.pa_source = pa_source if pa_source else guess_source()

        # rtp-rx-sink
        self.module_rtp_rx_sink = self._load_pa_module(['module-null-sink',
                                            'sink_name=rtp-rx-sink', 
                                            'rate=8000', 
                                            'channels=1'])
        if verbose: print(f'module_rtp_rx_sink: {self.module_rtp_rx_sink}')
        # rtp-tx-sink
        self.module_rtp_tx_sink = self._load_pa_module(['module-null-sink',
                                            'sink_name=rtp-tx-sink',
                                            'rate=8000',
                                            'channels=1'])
        if verbose: print(f'module_rtp_tx_sink: {self.module_rtp_tx_sink}')
        # loopback rx
        self.module_loopback_rx = self._load_pa_module(['module-loopback',
                                            'source=rtp-rx-sink.monitor',
                                            f'sink={self.pa_sink}',
                                            'latency_msec=70',
                                            'max_latency_msec=100',
                                            'fast_adjust_threshold_msec=100',
                                            'adjust_time=1',
                                            'rate=8000',
                                            'channels=1'])
        if verbose: print(f'module_loopback_rx: {self.module_loopback_rx}')
        # loopback tx
        self.module_loopback_tx = self._load_pa_module(['module-loopback',
                                            f'source={self.pa_source}',
                                            'sink=rtp-tx-sink',
                                            'latency_msec=70',
                                            'max_latency_msec=100',
                                            'fast_adjust_threshold_msec=100',
                                            'adjust_time=1',
                                            'rate=8000',
                                            'channels=1'])
        if verbose: print(f'module_loopback_tx: {self.module_loopback_tx}')
        # rtp-recv      Modules loaded in separate method to allow for easy
        # rtp-send      restart of RTP.
        self._rtp_load_modules()

        # create pipe to pacmd to communicate with pulseaudio
        # Stdout/stderr must be DEVNULL if running as systemd service.
        #       (This could be done more elegant with
        #       module-cli-protocol-unix.)
        self.pacmd = sp.Popen(['pacmd'], text=True,
                        stdin=sp.PIPE, stdout=sp.DEVNULL, stderr=sp.DEVNULL)

        # Set up volume
        # pyhsical sink (speaker): 100 %
        self._send_pacmd('set-sink-volume '
                         f'{self.pa_sink} {self._scale_volume(100)}')
        # rtp-rx-sink: configured volume
        self._send_pacmd('set-sink-volume rtp-rx-sink '
                         f'{self._scale_volume(pa_sink_volume)}')
        # rtp-tx-sink: 100 % (no volume processing)
        self._send_pacmd('set-sink-volume rtp-tx-sink '
                         f'{self._scale_volume(100)}')
        # physical source (mic): configured volume
        self._send_pacmd('set-source-volume '
                         f'{self.pa_source} {self._scale_volume(pa_source_volume)}')
        # unmute pysical source (mic)
        self._send_pacmd(f'set-source-mute {self.pa_source} 0')
        # unmute pysical sink (speaker)
        self._send_pacmd(f'set-sink-mute {self.pa_sink} 0')


    def done(self):
        # remove pulseaudio modules in reverse order of loading
        if verbose: print('Remove pulseaudio modules')
        self._rtp_remove_modules()
        self._unload_pa_module(self.module_loopback_tx)
        self._unload_pa_module(self.module_loopback_rx)
        self._unload_pa_module(self.module_rtp_tx_sink)
        self._unload_pa_module(self.module_rtp_rx_sink)
        # stop pacmd
        if verbose: print('Stop pacmd')
        self.pacmd.stdin.close()
        self.pacmd.terminate()
        self.pacmd.wait(timeout=1)
        # stop pulseaudio if we started it
        if not self.pa_was_running:
            if verbose: print('Stop Pulseaudio')
            sp.run(['pulseaudio', '--kill'], check=False)

    # --- private methods ----------------------------------------------------

    def _rtp_load_modules(self):
        """Load module-rtp-recv and module-rtp-send.

        The loading of these modules is separated from constuctor to implement
        an easy reload of these modules.
        """
        # rtp-recv
        self.module_rtp_recv = self._load_pa_module(['module-rtp-recv',
                                            'latency_msec=100',
                                            'sink=rtp-rx-sink',
                                            f'sap_address={self.rtp_mcast_ip}'])
        if verbose: print(f'module_rtp_recv: {self.module_rtp_recv}')
        # rtp-send
        self.module_rtp_send = self._load_pa_module(['module-rtp-send',
                                            'rate=8000',
                                            # ulaw geht auf bullsye nicht mehr.
                                            #'format=ulaw',
                                            'channels=1',
                                            'loop=false',
                                            'source=rtp-tx-sink.monitor',
                                            'inhibit_auto_suspend=always',
                                            f'destination={self.rtp_mcast_ip}'])
        if verbose: print(f'module_rtp_send: {self.module_rtp_send}')


    def _rtp_remove_modules(self):
        """Remove module-rtp-recv and module-rtp-send.

        The removing of these modules is separated from done() to
        implement an easy reload of these modules.
        """
        self._unload_pa_module(self.module_rtp_recv)
        self._unload_pa_module(self.module_rtp_send)

    # --- pulseaudio communication -------------------------------------------

    def _load_pa_module(self, args):
        """Load a pulseaudio module.

        args : list of strings
            arguments given to pactl

        Returns number of module as string.
        """
        # The modules are loaded using separate calls of pactl. The
        # interactive pacmd interface is not used, because it is difficult to
        # get the module numbers.
        args = ['pactl', 'load-module'] + args
        exit_status = sp.run(args, capture_output=True, check=True, text=True)
        number = exit_status.stdout.rstrip()
        return number

    def _unload_pa_module(self, n):
        "Unload module number n"
        # Module unloading is done in separate pactl calls. Depending on the
        # way of program termination, pacmd may be not running any more.
        # Even the pulseaudio daemon may be stopped, so do not check exit code.
        sp.run(['pactl', 'unload-module', str(n)], check=False)

    def _send_pacmd(self, cmd):
        """Send cmd to pulseaudio."""
        # The interactive pacmd interface is used for settings during
        # operation. Separate calls of pactl are too slow (esp. for volume
        # changes if the knop is turned fast).
        if verbose: print(cmd)
        self.pacmd.stdin.write(cmd + '\n')
        self.pacmd.stdin.flush()

    def _scale_volume(self, vol):
        "Scales volume value from [0..100] to [0..65536]"
        return int(vol / 100 * 65536)

    # --- public methods -----------------------------------------------------

    def rx_mute(self, mute):
        """Mute/unmute rx path"""
        self._send_pacmd(f'set-source-mute rtp-rx-sink.monitor {int(mute)}')

    def tx_mute(self, mute):
        """Mute/unmute tx path"""
        self._send_pacmd(f'set-source-mute rtp-tx-sink.monitor {int(mute)}')

    def set_volume(self, vol):
        """Set volume of rx path (speaker volume) in percent."""
        self._send_pacmd('set-sink-volume rtp-rx-sink '
                         f'{self._scale_volume(vol)}')

    def restart_rtp(self):
        """Remove and then load RTP modules from pulseaudio."""
        self._rtp_remove_modules()
        self._rtp_load_modules()

# ----------------------------------------------------------------------------
# guess sink and source for pulseaudio
# ----------------------------------------------------------------------------

def guess_sink():
    # The raspberry's sink is part of the broadcom chip, thus it has the
    # string 'bcm' in its name. Hopefully the other sink has not…
    sink = read_one_line_from_cmd("pactl list sinks short "
                                  "| grep -v bcm "
                                  "| cut -f 2")
    if verbose: print(f"Guessing sink: '{sink}'")
    return sink

def guess_source():
    # There should be only one _real_ source (i. e., a source that is not a
    # monitor of a sink).
    source = read_one_line_from_cmd("pactl list sources short "
                                    "| grep -v monitor "
                                    "| cut -f 2")
    if verbose: print(f"Guessing source: '{source}'")
    return source

def read_one_line_from_cmd(cmd):
    completed = sp.run(cmd,
                       capture_output=True,
                       shell=True,
                       check=True,
                       text=True)
    line = completed.stdout.strip()
    if '\n' in line:
        raise RuntimeError(f"Cmd '{cmd}' returned more than one line: "
                           f"'{line}'")
    if line == "":
        raise RuntimeError(f"Cmd '{cmd}' returned no output.")
    return line

# vim: set et fdm=marker:
